__all__ = [
    'Notification',
    'TriggerType',
    'check_witness',
    'notify',
    'log',
    'get_trigger',
    'get_notifications',
    'get_network',
    'burn_gas',
    'get_random',
    'load_script',
    'address_version',
    'executing_script_hash',
    'calling_script_hash',
    'time',
    'gas_left',
    'platform',
    'invocation_counter',
    'entry_script_hash',
    'script_container',
]

from collections.abc import Sequence
from typing import Any

from boa3.builtin.interop.blockchain import Transaction
from boa3.builtin.interop.contract.callflagstype import CallFlags
from boa3.builtin.interop.runtime.notification import Notification
from boa3.builtin.interop.runtime.triggertype import TriggerType
from boa3.builtin.type import ECPoint, UInt160
from boa3.internal.deprecation import deprecated


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def check_witness(hash_or_pubkey: UInt160 | ECPoint) -> bool:
    """
    Verifies that the transactions or block of the calling contract has validated the required script hash.

    >>> check_witness(calling_script_hash)
    True

    >>> check_witness(UInt160(bytes(20)))
    False

    :param hash_or_pubkey: script hash or public key to validate
    :type hash_or_pubkey: boa3.builtin.type.UInt160 or boa3.builtin.type.ECPoint
    :return: a boolean value that represents whether the script hash was verified
    :rtype: bool
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def notify(state: Any, notification_name: str = None):
    """
    Notifies the client from the executing smart contract.

    >>> var = 10
    ... notify(var)     # An event will be triggered
    None

    >>> var = 10
    ... notify(var, 'custom event name')     # An event will be triggered
    None

    :param state: the notification message
    :type state: Any
    :param notification_name: name that'll be linked to the notification
    :type notification_name: str
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def log(message: str):
    """
    Show log messages to the client from the executing smart contract.

    >>> log('log sent')     # An event that can be shown on the CLI
    log sent

    :param message: the log message
    :type message: str
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def get_trigger() -> TriggerType:
    """
    Return the smart contract trigger type.

    >>> get_trigger()
    TriggerType.APPLICATION

    :return: a value that represents the contract trigger type
    :rtype: boa3.builtin.interop.runtime.triggertype.TriggerType
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def get_notifications(script_hash: UInt160 = UInt160()) -> list[Notification]:
    """
    This method gets current invocation notifications from specific 'script_hash'.

    >>> notify(1); notify(2); notify(3)
    ... get_notifications(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))
    [
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [1]
        ],
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [2]
        ],
        [
            b'8\\xfe\\x11\\n\\xff7J\\xb8}\\xe9x6@\\xea\\x0b\\x00\\xf1|\\x82v',
            'notify',
            [3]
        ]
    ]

    :param script_hash: must have 20 bytes, but if it's all zero 0000...0000 it refers to all existing notifications
        (like a * wildcard)
    :type script_hash: boa3.builtin.type.UInt160
    :return: It will return an array of all matched notifications
    :rtype: list[boa3.builtin.interop.runtime.notification.Notification]
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def get_network() -> int:
    """
    Gets the magic number of the current network.

    >>> get_network()
    860243278

    :return: the magic number of the current network
    :rtype: int
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def burn_gas(gas: int):
    """
    Burns GAS to benefit the NEO ecosystem.

    >>> burn_gas(1000)
    None

    :param gas: the amount of GAS that will be burned
    :type gas: int

    :raise Exception: raised if gas value is negative.
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def get_random() -> int:
    """
    Gets the next random number.

    >>> get_random()
    191320526825634396960813166838892720709

    >>> get_random()
    99083669484001682562631729023191545809

    >>> get_random()
    328056213623902365838800581788496514419

    :return: the next random number
    :rtype: int
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.runtime` instead')
def load_script(script: bytes, args: Sequence = (), flags: CallFlags = CallFlags.NONE) -> Any:
    """
    Loads a script at runtime.

    >>> from typing import cast
    ... from boa3.builtin.vm import Opcode
    ... cast(int, load_script(Opcode.ADD, [10, 2]))
    12

    """
    pass


address_version: int = 0
"""
Gets the address version of the current network.

>>> address_version
53

:meta hide-value:
"""


executing_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the current context.

>>> executing_script_hash
b'^b]\\x90#\\xbf\\xcc\\x1f\\xd8\\x9e\\xe3\\xa4zd\\x14\\xa4\\xf0\\x96\\x9f`'

:meta hide-value:
"""

calling_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the calling contract.

>>> calling_script_hash
b'\\x05\\x7f\\xc2\\x9d\\xba\\xb2\\xc1x\\xf5\\x81\\x83\\xbf\\xcb\\x87/\\xc3!\\xca\\xe1\\xd0'

:meta hide-value:
"""

time: int = 0
"""
Gets the timestamp of the current block.

>>> time
1685395697108

:meta hide-value:
"""

gas_left: int = 0
"""
Gets the remaining GAS that can be spent in order to complete the execution.

>>> gas_left
1999015490

:meta hide-value:
"""

platform: str = ''
"""
Gets the name of the current platform.

>>> platform
'NEO'

:meta hide-value:
"""

invocation_counter: int = 0
"""
Gets the number of times the current contract has been called during the execution.

>>> invocation_counter
1

:meta hide-value:
"""

entry_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the entry context.

>>> entry_script_hash
b'\\tK\\xb31\\xa8\\x13\\x80`\\xad\\xf6\\xda\\xdf\\xc6R\\x9b\\xfdB\\xbf\\x83\\x8f'

:meta hide-value:
"""

script_container: Transaction = Transaction()
"""
Gets the current script container.

>>> script_container
[
    b'\\xf1y\\xc2\\xd6\\x1c\\xb6\\x98\\xa4\\xdc\\xf3\\xd67s\\xd7E\\xf0<;\\x98+\\xa2T\\x03P,T\\xe8\\xc6{ \\x101', 
    0, 
    442907905, 
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00', 
    0, 
    0, 
    0, 
    ''
]

>>> script_container
[
    b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
    0,
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
    '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
    1468595301000,
    2083236893,
    0,
    0,
    b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
    0,
]

:meta hide-value:
"""



from boa3.internal.model.builtin.interop.interop import Interop
from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type


class StrBytesGreaterThanOrEqual(BinaryOperation):
    """
    A class used to represent a string/bytes greater than or equal comparison

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.bytes, Type.str]

    def __init__(self, left: IType = Type.str, right: IType = None):
        self.operator: Operator = Operator.GtE
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left == right and any(_type.is_type_of(left) for _type in self._valid_types)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return Type.bool
        else:
            return Type.none

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp

        # comparing strings, on the stack the first string is the right one and the second is the left one
        Interop.MemoryCompare.generate_internal_opcodes(code_generator)
        # -1 means the left string is greater than the right string
        # 1 means the left string is less than the right string
        # 0 means the right string is equals to the left string

        # if comparison equals 1 (the left string is less than the right string), then return False
        code_generator.convert_literal(1)
        code_generator.convert_operation(BinaryOp.NumNotEq)
        # else, return True



__all__ = [
    'Block',
    'Signer',
    'Transaction',
    'VMState',
    'get_contract',
    'get_block',
    'get_transaction',
    'get_transaction_from_block',
    'get_transaction_height',
    'get_transaction_signers',
    'get_transaction_vm_state',
    'current_hash',
    'current_index',
]

from boa3.builtin.interop.blockchain.block import Block
from boa3.builtin.interop.blockchain.signer import Signer
from boa3.builtin.interop.blockchain.transaction import Transaction
from boa3.builtin.interop.blockchain.vmstate import VMState
from boa3.builtin.interop.contract import Contract
from boa3.builtin.type import UInt160, UInt256
from boa3.internal.deprecation import deprecated


@deprecated(details='This module is deprecated. Use :class:`ContractManagement` from :mod:`boa3.sc.contracts` instead')
def get_contract(hash: UInt160) -> Contract | None:
    """
    Gets a contract with a given hash. If the script hash is not associated with a smart contract, then it will return
    None.

    >>> get_contract(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))    # GAS script hash
    {
        'id': -6,
        'update_counter': 0,
        'hash': b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2',
        'nef': b'NEF3neo-core-v3.0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00#\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@QA\\xc7\\x9e',
        'manifest': {
            'name': 'GasToken',
            'group': [],
            'supported_standards': ['NEP-17'],
            'abi': [[['balanceOf', [['account', 20]], 17, 0, True], ['decimals', [], 17, 7, True], ['symbol', [], 19, 14, True], ['totalSupply', [], 17, 21, True], ['transfer', [['from', 20], ['to', 20], ['amount', 17], ['data', 0]], 16, 28, False]], [['Transfer', [['from', 20], ['to', 20], ['amount', 17]]]]],
            'permissions': [[None, None]],
            'trusts': [],
            'extras': 'null'
        },
    }

    >>> get_contract(UInt160(bytes(20)))    # there is no smart contract associated with this script hash
    None

    :param hash: a smart contract hash
    :type hash: boa3.builtin.type.UInt160
    :return: a contract
    :rtype: boa3.builtin.interop.contract.contract.Contract

    :raise Exception: raised if hash length isn't 20 bytes.
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_block(index_or_hash: int | UInt256) -> Block | None:
    """
    Gets the block with the given index or hash. Will return None if the index or hash is not associated with a Block.

    >>> get_block(0)        # first block
    {
        'hash': b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
        'version': 0,
        'previous_hash': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
        'merkle_root': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
        'timestamp': 1468595301000,
        'nonce': 2083236893,
        'index': 0,
        'primary_index': 0,
        'next_consensus': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
        'transaction_count': 0,
    }

    >>> get_block(UInt256(b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04"))        # first block
    {
        'hash': b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
        'version': 0,
        'previous_hash': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
        'merkle_root': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
        'timestamp': 1468595301000,
        'nonce': 2083236893,
        'index': 0,
        'primary_index': 0,
        'next_consensus': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
        'transaction_count': 0,
    }

    >>> get_block(9999999)      # block doesn't exist
    None

    >>> get_block(UInt256(bytes(32)))   # block doesn't exist
    None

    :param index_or_hash: index or hash identifier of the block
    :type index_or_hash: int or boa3.builtin.type.UInt256
    :return: the desired block, if exists. None otherwise
    :rtype: boa3.builtin.interop.blockchain.block.Block or None
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_transaction(hash_: UInt256) -> Transaction | None:
    """
    Gets a transaction with the given hash. Will return None if the hash is not associated with a Transaction.

    >>> get_transaction(UInt256(b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr'))
    {
        'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
        'version': 0,
        'nonce': 2025056010,
        'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
        'system_fee': 2028330,
        'network_fee': 1206580,
        'valid_until_block': 5761,
        'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
    }

    >>> get_transaction(UInt256(bytes(32)))   # transaction doesn't exist
    None

    :param hash_: hash identifier of the transaction
    :type hash_: boa3.builtin.type.UInt256
    :return: the Transaction, if exists. None otherwise
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_transaction_from_block(block_hash_or_height: UInt256 | int, tx_index: int) -> Transaction | None:
    """
    Gets a transaction from a block. Will return None if the block hash or height is not associated with a Transaction.

    >>> get_transaction_from_block(1, 0)
    {
        'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
        'version': 0,
        'nonce': 2025056010,
        'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
        'system_fee': 2028330,
        'network_fee': 1206580,
        'valid_until_block': 5761,
        'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
    }

    >>> get_transaction_from_block(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'), 0)
    {
        'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
        'version': 0,
        'nonce': 2025056010,
        'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
        'system_fee': 2028330,
        'network_fee': 1206580,
        'valid_until_block': 5761,
        'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
    }

    >>> get_transaction_from_block(123456789, 0)     # height does not exist yet
    None

    >>> get_transaction_from_block(UInt256(bytes(32)), 0)     # block hash does not exist
    None

    :param block_hash_or_height: a block identifier
    :type block_hash_or_height: boa3.builtin.type.UInt256 or int
    :param tx_index: the transaction identifier in the block
    :type tx_index: int
    :return: the Transaction, if exists. None otherwise
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_transaction_height(hash_: UInt256) -> int:
    """
    Gets the height of a transaction. Will return -1 if the hash is not associated with a Transaction.

    >>> get_transaction_height(UInt256(b'\\x28\\x89\\x4f\\xb6\\x10\\x62\\x9d\\xea\\x4c\\xcd\\x00\\x2e\\x9e\\x11\\xa6\\xd0\\x3d\\x28\\x90\\xc0\\xe5\\xd4\\xfc\\x8f\\xc6\\x4f\\xcc\\x32\\x53\\xb5\\x48\\x01'))
    2108703

    >>> get_transaction_height(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
    10

    >>> get_transaction_height(UInt256(bytes(32)))   # transaction doesn't exist
    -1

    :param hash_: hash identifier of the transaction
    :type hash_: boa3.builtin.type.UInt256
    :return: height of the transaction
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_transaction_signers(hash_: UInt256) -> list[Signer]:
    """
    Gets a list with the signers of a transaction.

    >>> get_transaction_signers(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
    [
        {
            "account": b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            "scopes": 1,
            "allowed_contracts": [],
            "allowed_groups": [],
            "rules": [],
        },
    ]

    :param hash_: hash identifier of the transaction
    :type hash_: boa3.builtin.type.UInt256
    :return: VM state of the transaction
    """
    pass


@deprecated(details='This module is deprecated. Use :class:`LedgerContract` from :mod:`boa3.sc.contracts` instead')
def get_transaction_vm_state(hash_: UInt256) -> VMState:
    """
    Gets the VM state of a transaction.

    >>> get_transaction_vm_state(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
    VMState.HALT

    :param hash_: hash identifier of the transaction
    :type hash_: boa3.builtin.type.UInt256
    :return: VM state of the transaction
    """
    pass


current_hash: UInt256 = UInt256()
"""
Gets the hash of the current block.

>>> current_hash
b'\\x3e\\x65\\xe5\\x4d\\x75\\x5a\\x94\\x90\\xd6\\x98\\x3a\\x77\\xe4\\x82\\xaf\\x7a\\x38\\xc9\\x8c\\x1a\\xc6\\xd9\\xda\\x48\\xbd\\x7c\\x22\\xb3\\x2a\\x9e\\x34\\xea'

:meta hide-value:
"""

current_index: int = 0
"""
Gets the index of the current block.

>>> current_index
10908937

>>> current_index
2108690

>>> current_index
3529755

:meta hide-value:
"""



from boa3.internal.model.builtin.interop.nativecontract import PolicyContractMethod
from boa3.internal.model.builtin.interop.policy.transactionattributetype import TransactionAttributeType
from boa3.internal.model.variable import Variable


class SetAttributeFeeMethod(PolicyContractMethod):

    def __init__(self, tx_attribute_type: TransactionAttributeType):
        from boa3.internal.model.type.type import Type
        identifier = 'set_attribute_fee'
        native_identifier = 'setAttributeFee'
        args: dict[str, Variable] = {
            'attribute_type': Variable(tx_attribute_type),
            'value': Variable(Type.int),
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.none)



import abc
import ast
from collections.abc import Iterable, Sized
from typing import Any

from boa3.internal.model import set_internal_call
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class IStoragePutMethod(InteropMethod, abc.ABC):

    def __init__(self, identifier: str, value_type: IType):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.builtin.interop.storage.storagecontext.storagecontexttype import StorageContextType

        syscall = 'System.Storage.Put'
        context_type = StorageContextType.build()

        args: dict[str, Variable] = {'key': Variable(Type.bytes),
                                     'value': Variable(value_type),
                                     'context': Variable(context_type)}

        from boa3.internal.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
        default_id = StorageGetContextMethod(context_type).identifier
        context_default = set_internal_call(ast.parse(f"{default_id}()"
                                                      ).body[0].value)
        super().__init__(identifier, syscall, args, defaults=[context_default], return_type=Type.none)

    @abc.abstractmethod
    def generate_serialize_value_opcodes(self, code_generator):
        pass

    def generate_internal_opcodes(self, code_generator):
        start_address = code_generator.bytecode_size
        code_generator.swap_reverse_stack_items(3)
        self.generate_serialize_value_opcodes(code_generator)
        end_address = code_generator.last_code_start_address

        if end_address > start_address:
            code_generator.swap_reverse_stack_items(3)
        else:
            code_generator._remove_inserted_opcodes_since(start_address)

        super().generate_internal_opcodes(code_generator)

    @property
    def generation_order(self) -> list[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = super().generation_order
        context_index = list(self.args).index('context')

        if indexes[-1] != context_index:
            # context must be the last generated argument
            indexes.remove(context_index)
            indexes.append(context_index)

        return indexes

    @property
    def key_arg(self) -> Variable:
        return self.args['key']

    @property
    def value_arg(self) -> Variable:
        return self.args['value']

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, (Sized, Iterable)):
            return self
        num_args: int = len(self.args)
        if len(value) != num_args or any(not isinstance(exp, (IExpression, IType)) for exp in value[:num_args]):
            return self

        exp = [exp if isinstance(exp, IExpression) else Variable(exp) for exp in value]
        if not self.validate_parameters(*exp):
            return self

        key_type: IType = exp[0].type
        value_type: IType = exp[1].type
        if self.key_arg.type.is_type_of(key_type) and self.value_arg.type.is_type_of(value_type):
            return self

        from boa3.internal.model.builtin.interop.storage.put.storageputbytesmethod import StoragePutBytesMethod
        method: InteropMethod = StoragePutBytesMethod()
        method.args['key'] = Variable(key_type)
        method.args['value'] = Variable(value_type)
        return method



from __future__ import annotations

__all__ = ['StorageContext']

from boa3.builtin.interop.storage.storagemap import StorageMap
from boa3.internal.deprecation import deprecated


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.storage` instead')
class StorageContext:
    """
    The storage context used to read and write data in smart contracts.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/services/StorageContext>`__
    to learn more about the StorageContext class.
    """

    def __init__(self):
        pass

    def create_map(self, prefix: bytes) -> StorageMap:
        """
        Creates a storage map with the given prefix.

        :param prefix: the identifier of the storage map
        :type prefix: bytes
        :return: a map with the key-values in the storage that match with the given prefix
        :rtype: boa3.builtin.interop.storage.storagemap.StorageMap
        """
        pass

    def as_read_only(self) -> StorageContext:
        """
        Converts the specified storage context to a new readonly storage context.

        :return: current StorageContext as ReadOnly
        :rtype: StorageContext
        """
        pass



from boa3.internal.model.builtin.interop.storage.get.istoragegetmethod import IStorageGetMethod


class StorageGetBoolMethod(IStorageGetMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'get_bool'
        value_type = Type.bool

        super().__init__(identifier, value_type=value_type)

    def generate_default_value_opcodes(self, code_generator):
        # default_value = 0
        code_generator.convert_literal(0)

    def generate_deserialize_value_opcodes(self, code_generator):
        from boa3.internal.model.builtin.method.toboolmethod import ToBool
        from boa3.internal.model.type.type import Type

        converter = ToBool.build(Type.bytes)
        code_generator.convert_builtin_method_call(converter)



from typing import Any

from boa3.builtin.interop.contract import Contract
from boa3.builtin.type import UInt160
from boa3.internal.deprecation import deprecated


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.types` instead')
class Nep17Contract(Contract):
    def symbol(self) -> str:
        pass

    def decimals(self) -> int:
        pass

    def total_supply(self) -> int:
        pass

    def balance_of(self, account: UInt160) -> int:
        pass

    def transfer(self, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        pass



__all__ = [
    'CreateNewEvent',
    'public',
    'contract',
    'display_name',
    'NeoMetadata',
]

from boa3.builtin.type import Event
from boa3.internal.deprecation import deprecated
from boa3.sc.compiletime import NeoMetadata


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.utils` instead')
def CreateNewEvent(arguments: list[tuple[str, type]] = [], event_name: str = '') -> Event:
    """
    Creates a new Event.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/basics#events>`__ to learn more
    about Events.

    >>> new_event: Event = CreateNewEvent(
    ...     [
    ...        ('name', str),
    ...        ('amount', int)
    ...     ],
    ...     'New Event'
    ... )

    :param arguments: the list of the events args' names and types
    :type arguments: list[tuple[str, type]]
    :param event_name: custom name of the event. It's filled with the variable name if not specified
    :type event_name: str
    :return: the new event
    :rtype: boa3.builtin.type.Event
    """
    pass


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.compiletime` instead')
def public(name: str = None, safe: bool = False, *args, **kwargs):
    """
    This decorator identifies which methods should be included in the abi file. Adding this decorator to a function
    means it could be called externally.

    >>> @public     # this method will be added to the abi
    ... def callable_function() -> bool:
    ...     return True
    {
        "name": "callable_function",
        "offset": 0,
        "parameters": [],
        "safe": false,
        "returntype": "Boolean"
    }

    >>> @public(name='callableFunction')     # the method will be added with the different name to the abi
    ... def callable_function() -> bool:
    ...     return True
    {
        "name": "callableFunction",
        "offset": 0,
        "parameters": [],
        "safe": false,
        "returntype": "Boolean"
    }

    >>> @public(safe=True)      # the method will be added with the safe flag to the abi
    ... def callable_function() -> bool:
    ...     return True
    {
        "name": "callable_function",
        "offset": 0,
        "parameters": [],
        "safe": true,
        "returntype": "Boolean"
    }

    :param name: Identifier for this method that'll be used on the abi. If not specified, it'll be the same
     identifier from Python method definition
    :type name: str
    :param safe: Whether this method is an abi safe method. False by default
    :type safe: bool
    """
    def decorator_wrapper(*args, **kwargs):
        pass
    return decorator_wrapper


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.compiletime` instead')
def contract(script_hash: str | bytes):
    """
    This decorator identifies a class that should be interpreted as an interface to an existing contract.

    If you want to use the script hash in your code, you can use the `hash` class attribute that automatically maps the
    script hash parameter onto it. You don't need to declare it in your class, but your IDE might send a warning about
    the attribute if you don't.

    Check out `Our Documentation <https://dojo.coz.io/neo3/boa/calling-smart-contracts.html#with-interface>`__ to learn
    more about this decorator.

    >>> @contract('0xd2a4cff31913016155e38e474a2c06d08be276cf')
    ... class GASInterface:
    ...     hash: UInt160      # you don't need to declare this class variable, we are only doing it to avoid IDE warnings
    ...                        # but if you do declare, you need to import the type UInt160 from boa3.builtin.type
    ...     @staticmethod
    ...     def symbol() -> str:
    ...         pass
    ... @public
    ... def main() -> str:
    ...     return "Symbol is " + GASInterface.symbol()
    ... @public
    ... def return_hash() -> UInt160:
    ...     return GASInterface.hash    # neo3-boa will understand that this attribute exists even if you don't declare it

    :param script_hash: Script hash of the interfaced contract
    :type script_hash: str or bytes
    """
    def decorator_wrapper(cls, *args, **kwargs):
        if isinstance(script_hash, str):
            from boa3.internal.neo import from_hex_str
            _hash = from_hex_str(script_hash)
        else:
            _hash = script_hash

        cls.hash = _hash
        return cls
    return decorator_wrapper


@deprecated(details='This module is deprecated. Use :mod:`boa3.sc.compiletime` instead')
def display_name(name: str):
    """
    This decorator identifies which methods from a contract interface should have a different identifier from the one
    interfacing it. It only works in contract interface classes.

    >>> @contract('0xd2a4cff31913016155e38e474a2c06d08be276cf')
    ... class GASInterface:
    ...     @staticmethod
    ...     @display_name('totalSupply')
    ...     def total_supply() -> int:      # the smart contract will call "totalSupply", but when writing the script you can call this method whatever you want to
    ...         pass
    ... @public
    ... def main() -> int:
    ...     return GASInterface.total_supply()

    :param name: Method identifier from the contract manifest.
    :type name: str
    """
    def decorator_wrapper(*args, **kwargs):
        pass
    return decorator_wrapper



from typing import Any

from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.itype import IType


class ScriptHashType(UInt160Type):
    """
    A class used to indicate that a parameter or return on the manifest is a Scripthash.
    It's a subclass of UInt160Type.
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'ScriptHash'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _ScriptHash


_ScriptHash = ScriptHashType()



from boa3.internal.model.builtin.classmethod.copymethod import CopyMethod
from boa3.internal.model.type.itype import IType


class CopyListMethod(CopyMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        super().__init__(arg_value if Type.list.is_type_of(arg_value) else Type.list)
